// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_DIGITNUMBER
#define FF_ST_DIGITNUMBER

#define LOG log10(2)

template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline int flexfloat_st<exp_bits, frac_bits>::nb_significant_digit() const
{
#ifdef FF_ST_FLOAT128
    fesetround(FE_TONEAREST); /// temporaire
    const uint_fast8_t MAX_PREC = (uint_fast8_t)((double)(frac_bits + 1) * LOG);

    __float128 x0, x1, x2, xx;

    xx = x.getValue() + y.getValue() + z.getValue();

    accuracy = 0;
    if (xx == 0.0)
    {
        if ((x == y) && (x == z))
            accuracy = MAX_PREC;
    }
    else
    {
        xx = (__float128)3. / xx;
        x0 = x.getValue() * xx - (__float128)1.;
        x1 = y.getValue() * xx - (__float128)1.;
        x2 = z.getValue() * xx - (__float128)1.;

        if (isnanq(xx) || isinfq(xx))
        {
            accuracy = 0;
        }
        else
        {
            __float128 yy = (x0 * x0 + x1 * x1 + x2 * x2) * 3.085466170416667;
            if (yy <= 1.e-68)
                accuracy = MAX_PREC;
            else
            {
                __int128_t e = ((float128bf *)&yy)->bf;
                e = (((float128bf *)&yy)->bf) & (((__uint128_t)0x7FFF) << 112);
                e >>= 112;
                e -= 16383;
                yy = -e * 0.301029995663981195;
                if (yy >= (__float128)0.)
                    accuracy = (int)((yy + (__float128)1.) * (__float128)0.5);
                // la formule varie légèrement est-ce normal ? acc = (yy + 1)/2 au lieu de acc = yy + 1/2
            }
        }
    }

#else
    fesetround(FE_TONEAREST); /// temporaire
    const uint_fast8_t MAX_PREC = (uint_fast8_t)((double)(frac_bits + 1) * LOG);

    double x0, x1, x2, xx;

    xx = x.getValue() + y.getValue() + z.getValue();

    accuracy = 0;
    if (xx == 0.0)
    {
        if (x == y && x == z)
            accuracy = MAX_PREC;
    }
    else
    {
        xx = (double)3. / xx;
        x0 = x.getValue() * xx - (double)1.;
        x1 = y.getValue() * xx - (double)1.;
        x2 = z.getValue() * xx - (double)1.;

        if (std::isnan(xx) || std::isinf(xx))
        {
            accuracy = 0;
        }
        else
        {
            double yy = (x0 * x0 + x1 * x1 + x2 * x2) * 3.085466170416667;
            if (yy <= 1.e-30)
                accuracy = MAX_PREC;
            else
            {
                long e = ((doublebf *)&yy)->bf;
                e &= 0x7FF0000000000000UL;
                e >>= 52;
                e -= 1021;
                yy = -e * 0.15051499783199065;
                if (yy >= 0.)
                    accuracy = (uint_fast8_t)(yy + 0.5);
            }
        }
    }
#endif

    fesetround(FE_UPWARD);
    return accuracy;
}
template <uint_fast8_t exp_bits, uint_fast8_t frac_bits>
inline int flexfloat_st<exp_bits, frac_bits>::approx_digit() const
{
    double x0, x1, x2, xx;

    accuracy = RELIABLE_RESULT;
    xx = x.getValue() + y.getValue() + z.getValue();

    if (std::isinf(xx) || std::isnan(xx))
        accuracy = 0;

    if (xx == (double)0.0)
    {
        if ((x.getValue() != y.getValue()) || (x.getValue() != z.getValue()))
            accuracy = 0;
    }
    else
    {
        xx = (double)3. / xx;
        x0 = x.getValue() * xx - (double)1.;
        x1 = y.getValue() * xx - (double)1.;
        x2 = z.getValue() * xx - (double)1.;
        if ((x0 * x0 + x1 * x1 + x2 * x2) > 3.241001342318910E-02)
            accuracy = 0;
    }

    return accuracy;
}

#endif