// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FF_ST_MATH_HPP
#define FF_ST_MATH_HPP

friend flexfloat_st<exp_bits, frac_bits> sqrt(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    res.x = sqrt((double)a.x.getValue());
    res.y = sqrt((double)a.y.getValue());
    res.z = sqrt((double)a.z.getValue());
    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b.x;
    else
        res.x = -((-a.x) - (-b.x));

    if (RANDOM)
    {
        res.y = a.y - b.y;
        res.z = -((-a.z) - (-b.z));
    }
    else
    {
        res.z = a.z - b.z;
        res.y = -((-a.y) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED || a.accuracy == RELIABLE_RESULT)
            a.nb_significant_digit();
        if (b.accuracy == DIGIT_NOT_COMPUTED || b.accuracy == RELIABLE_RESULT)
            b.nb_significant_digit();
        if (a.accuracy > b.accuracy)
            res = a;
        else
            res = b;
    }
    else
    {
        if ((a.x + a.y + a.z) < (b.x + b.y + b.z))
            res = a;
        else
            res = b;
    }
    return res;
}


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> fmin(const __float128 &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> fmin(const long double &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const double &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const float &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const int &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend flexfloat_st<exp_bits, frac_bits> fmin(const half &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(a);
    else
    {
        if (3 * a < (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}
#endif


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const __float128 &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const long double &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const double &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const float &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const int &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend flexfloat_st<exp_bits, frac_bits> fmin(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const half &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) < 3 * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a,
                                              const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b.x;
    else
        res.x = -((-a.x) - (-b.x));

    if (RANDOM)
    {
        res.y = a.y - b.y;
        res.z = -((-a.z) - (-b.z));
    }
    else
    {
        res.z = a.z - b.z;
        res.y = -((-a.y) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED || a.accuracy == RELIABLE_RESULT)
            a.nb_significant_digit();
        if (b.accuracy == DIGIT_NOT_COMPUTED || b.accuracy == RELIABLE_RESULT)
            b.nb_significant_digit();
        if (a.accuracy > b.accuracy)
            res = a;
        else
            res = b;
    }
    else
    {
        if ((a.x + a.y + a.z) > (b.x + b.y + b.z))
            res = a;
        else
            res = b;
    }
    return res;
}


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> fmax(const __float128 &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> fmax(const long double &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const double &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const float &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const int &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend flexfloat_st<exp_bits, frac_bits> fmax(const half &a, const flexfloat_st<exp_bits, frac_bits> &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a - b.x;
    else
        res.x = -((-a) - (-b.x));

    if (RANDOM)
    {
        res.y = a - b.y;
        res.z = -((-a) - (-b.z));
    }
    else
    {
        res.z = a - b.z;
        res.y = -((-a) - (-b.y));
    }

    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);
    if (r)
    {
        res = flexfloat_st<exp_bits, frac_bits>(a);
    }
    else
    {
        if (3 * a > (b.x + b.y + b.z))
            res = flexfloat_st<exp_bits, frac_bits>(a);
        else
            res = b;
    }
    return res;
}
#endif


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const __float128 &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const long double &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const double &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const float &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const int &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}

#ifdef FF_ST_HALF_EMULATION
friend flexfloat_st<exp_bits, frac_bits> fmax(const flexfloat_st<exp_bits, frac_bits> &a, const half &b)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (RANDOM)
        res.x = a.x - b;
    else
        res.x = -((-a.x) - (-b));

    if (RANDOM)
    {
        res.y = a.y - b;
        res.z = -((-a.z) - (-b));
    }
    else
    {
        res.z = a.z - b;
        res.y = -((-a.y) - (-b));
    }
    int r = res.numericalnoise();

    if (_ff_st_branching_tag && r)
        instability(&_ff_st_branching_count);

    if (r)
        res = flexfloat_st<exp_bits, frac_bits>(b);
    else
    {
        if ((a.x + a.y + a.z) > 3. * b)
            res = a;
        else
            res = flexfloat_st<exp_bits, frac_bits>(b);
    }
    return res;
}
#endif

friend flexfloat_st<exp_bits, frac_bits> log(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = logq((__float128)a.x.getValue());
    res.y = logq((__float128)a.y.getValue());
    res.z = logq((__float128)a.z.getValue());
    #else
    res.x = log((double)a.x.getValue());
    res.y = log((double)a.y.getValue());
    res.z = log((double)a.z.getValue());
    #endif
    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> exp(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = expq((__float128)a.x.getValue());
    res.y = expq((__float128)a.y.getValue());
    res.z = expq((__float128)a.z.getValue());
    #else
    res.x = exp((double)a.x.getValue());
    res.y = exp((double)a.y.getValue());
    res.z = exp((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> cos(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = cosq((__float128)a.x.getValue());
    res.y = cosq((__float128)a.y.getValue());
    res.z = cosq((__float128)a.z.getValue());
    #else
    res.x = cos((double)a.x.getValue());
    res.y = cos((double)a.y.getValue());
    res.z = cos((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> sin(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = sinq((__float128)a.x.getValue());
    res.y = sinq((__float128)a.y.getValue());
    res.z = sinq((__float128)a.z.getValue());
    #else
    res.x = sin((double)a.x.getValue());
    res.y = sin((double)a.y.getValue());
    res.z = sin((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> tan(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = tanq((__float128)a.x.getValue());
    res.y = tanq((__float128)a.y.getValue());
    res.z = tanq((__float128)a.z.getValue());
    #else
    res.x = tan((double)a.x.getValue());
    res.y = tan((double)a.y.getValue());
    res.z = tan((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> acos(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = acosq((__float128)a.x.getValue());
    res.y = acosq((__float128)a.y.getValue());
    res.z = acosq((__float128)a.z.getValue());
    #else
    res.x = acos((double)a.x.getValue());
    res.y = acos((double)a.y.getValue());
    res.z = acos((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> asin(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = asinq((__float128)a.x.getValue());
    res.y = asinq((__float128)a.y.getValue());
    res.z = asinq((__float128)a.z.getValue());
    #else
    res.x = asin((double)a.x.getValue());
    res.y = asin((double)a.y.getValue());
    res.z = asin((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> atan(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = atanq((__float128)a.x.getValue());
    res.y = atanq((__float128)a.y.getValue());
    res.z = atanq((__float128)a.z.getValue());
    #else
    res.x = atan((double)a.x.getValue());
    res.y = atan((double)a.y.getValue());
    res.z = atan((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> cosh(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = coshq((__float128)a.x.getValue());
    res.y = coshq((__float128)a.y.getValue());
    res.z = coshq((__float128)a.z.getValue());
    #else
    res.x = cosh((double)a.x.getValue());
    res.y = cosh((double)a.y.getValue());
    res.z = cosh((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> sinh(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = sinhq((__float128)a.x.getValue());
    res.y = sinhq((__float128)a.y.getValue());
    res.z = sinhq((__float128)a.z.getValue());
    #else
    res.x = sinh((double)a.x.getValue());
    res.y = sinh((double)a.y.getValue());
    res.z = sinh((double)a.z.getValue());
    #endif
    fesetround(FE_UPWARD);

    return res;
};

friend flexfloat_st<exp_bits, frac_bits> tanh(const flexfloat_st<exp_bits, frac_bits> &a)
{
    flexfloat_st<exp_bits, frac_bits> res;

    if (_ff_st_math_tag != 0)
    {
        if (a.accuracy == DIGIT_NOT_COMPUTED)
            a.approx_digit();
        if (a.accuracy == 0)
            instability(&_ff_st_math_count);
    }
    res.accuracy = DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x = tanhq((__float128)a.x.getValue());
    res.y = tanhq((__float128)a.y.getValue());
    res.z = tanhq((__float128)a.z.getValue());
    #else
    res.x = tanh((double)a.x.getValue());
    res.y = tanh((double)a.y.getValue());
    res.z = tanh((double)a.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};



///////////////////// POW ///////////////////


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> pow(const __float128 &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    fesetround(FE_UPWARD);

    return res;
};
#endif


friend flexfloat_st<exp_bits, frac_bits> pow(const double &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const float &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif
    fesetround(FE_UPWARD);

    return res;
};


#ifdef FF_ST_HALF_EMULAION
friend flexfloat_st<exp_bits, frac_bits> pow(const half &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};
#endif



friend flexfloat_st<exp_bits, frac_bits> pow(const unsigned long long &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif
    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const long long &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const unsigned long &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const long &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const unsigned int &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const int &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const unsigned short &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const short &a, const flexfloat_st<exp_bits, frac_bits> &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (b.accuracy==DIGIT_NOT_COMPUTED) b.approx_digit();
        if (b.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a,(__float128)b.x.getValue());
    res.y=powq((__float128)a,(__float128)b.y.getValue());
    res.z=powq((__float128)a,(__float128)b.z.getValue());
    #else
    res.x=pow((double)a,(double)b.x.getValue());
    res.y=pow((double)a,(double)b.y.getValue());
    res.z=pow((double)a,(double)b.z.getValue());
    #endif

    fesetround(FE_UPWARD);

    return res;
};


#ifdef FF_ST_FLOAT128
friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const __float128 &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    fesetround(FE_UPWARD);

    return res;
};
#endif


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const double &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);

    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const float &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


#ifdef FF_ST_HALF_EMULATION
friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const half &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};
#endif


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const unsigned long long &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const long long &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const unsigned long &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const long &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const unsigned int &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const int &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const unsigned short &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif

    fesetround(FE_UPWARD);

    return res;
};


friend flexfloat_st<exp_bits, frac_bits> pow(const flexfloat_st<exp_bits, frac_bits> &a, const short &b) {
    flexfloat_st<exp_bits, frac_bits> res;

    if(_ff_st_power_tag!=0){
        if (a.accuracy==DIGIT_NOT_COMPUTED) a.approx_digit();
        if (a.accuracy==0) instability(&_ff_st_power_count);
    }

    res.accuracy=DIGIT_NOT_COMPUTED;
    fesetround(FE_TONEAREST);
    
    #ifdef FF_ST_FLOAT128
    res.x=powq((__float128)a.x.getValue(),(__float128)b);
    res.y=powq((__float128)a.y.getValue(),(__float128)b);
    res.z=powq((__float128)a.z.getValue(),(__float128)b);
    #else
    res.x=pow((double)a.x.getValue(),(double)b);
    res.y=pow((double)a.y.getValue(),(double)b);
    res.z=pow((double)a.z.getValue(),(double)b);
    #endif
    
    fesetround(FE_UPWARD);

    return res;
};


#endif