// Copyright 2024 Fabienne Jezequel, Ivan Lucas, Enzo Roaldes

// This file is part of SAFE.

// SAFE is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// SAFE is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with SAFE.  If not, see <http://www.gnu.org/licenses/>.

#include <cstdio>
#include <iostream>
#include "../include/flexfloat_st.hpp"

#define PACKAGE_STRING "SAFE"

///////////// INIT /////////////

void ff_st_init(int max_nb_instability)
{
    ff_st_init(max_nb_instability, 0, 51, _ff_st_cancel_value);
}

void ff_st_init(int max_nb_instability, unsigned int tag)
{
    ff_st_init(max_nb_instability, tag, 51, _ff_st_cancel_value);
}

void ff_st_init(int max_instability, unsigned int tag, unsigned int seed, unsigned int cancellation)
{
    _ff_st_cancel_value = cancellation;

    _ff_st_max_instability = max_instability;
    _ff_st_instability_detected = 0;

    _ff_st_div_count = 0;
    _ff_st_mul_count = 0;
    _ff_st_power_count = 0;
    _ff_st_math_count = 0;
    _ff_st_intrinsic_count = 0;
    _ff_st_cancel_count = 0;
    _ff_st_branching_count = 0;
    #ifdef FF_ST_HALF_EMULATION
    _ff_st_half_overflow_count=0; 
    _ff_st_half_underflow_count=0; 
    #endif

    if (max_instability == 0)
    {
        _ff_st_div_change = 1;
        _ff_st_mul_change = 1;
        _ff_st_power_change = 1;
        _ff_st_branching_change = 1;
        _ff_st_cancel_change = 1;
        _ff_st_intrinsic_change = 1;
        _ff_st_math_change = 1;
        #ifdef FF_ST_HALF_EMULATION
        _ff_st_half_overflow_change=1;
        _ff_st_half_underflow_change=1;
        #endif

        _ff_st_div_tag = 0;
        _ff_st_mul_tag = 0;
        _ff_st_power_tag = 0;
        _ff_st_branching_tag = 0;
        _ff_st_cancel_tag = 0;
        _ff_st_intrinsic_tag = 0;
        _ff_st_math_tag = 0;
        #ifdef FF_ST_HALF_EMULATION
        _ff_st_half_overflow_tag=0;
        _ff_st_half_underflow_tag=0;
        #endif
    }
    else
    {
        if (tag & FF_ST_DIV)
        {
            _ff_st_div_tag = 0;
            _ff_st_div_change = 1;
        }
        if (tag & FF_ST_MUL)
        {
            _ff_st_mul_tag = 0;
            _ff_st_mul_change = 1;
        }
        if (tag & FF_ST_POWER)
        {
            _ff_st_power_tag = 0;
            _ff_st_power_change = 1;
        }
        if (tag & FF_ST_MATH)
        {
            _ff_st_math_tag = 0;
            _ff_st_math_change = 1;
        }
        if (tag & FF_ST_INTRINSIC)
        {
            _ff_st_intrinsic_tag = 0;
            _ff_st_intrinsic_change = 1;
        }
        if (tag & FF_ST_CANCEL)
        {
            _ff_st_cancel_tag = 0;
            _ff_st_cancel_change = 1;
        }
        if (tag & FF_ST_BRANCHING)
        {
            _ff_st_branching_tag = 0;
            _ff_st_branching_change = 1;
        }
        #ifdef FF_ST_HALF_EMULATION
        if (tag & FF_ST_HALF_OVERFLOW) {
            _ff_st_half_overflow_tag=0; 
            _ff_st_half_overflow_change=1; 
        }
        if (tag & FF_ST_HALF_UNDERFLOW) {
            _ff_st_half_underflow_tag=0; 
            _ff_st_half_underflow_change=1; 
        }
        #endif
    }
    std::cout << "----------------------------------------------------------------" << std::endl;

    std::cout << PACKAGE_STRING << " software" << std::endl;

    if (_ff_st_div_tag & _ff_st_mul_tag & _ff_st_power_tag)
        std::cout << "Self-validation detection: ON" << std::endl;
    else
        std::cout << "Self-validation detection: OFF" << std::endl;
    std::cout << "Mathematical instabilities detection: " << (_ff_st_math_tag ? "ON" : "OFF") << std::endl;
    std::cout << "Branching instabilities detection: " << (_ff_st_branching_tag ? "ON" : "OFF") << std::endl;
    std::cout << "Intrinsic instabilities detection: " << (_ff_st_intrinsic_tag ? "ON" : "OFF") << std::endl;
    std::cout << "Cancellation instabilities detection: " << (_ff_st_cancel_tag ? "ON" : "OFF") << std::endl;
    #ifdef FF_ST_HALF_EMULATION
    std::cout << "Overflow detection for half-precision: " << (_ff_st_half_overflow_tag ? "ON": "OFF") << std::endl;
    std::cout << "Underflow detection for half-precision: " << (_ff_st_half_underflow_tag ? "ON": "OFF") << std::endl;
    #endif
    std::cout << "----------------------------------------------------------------" << std::endl;

    _st_init_table(4);
    fesetround(FE_UPWARD);

    _ff_st_recurrence = seed;
    _ff_st_random = 0;
    _ff_st_random_counter = 0;
}



///////////// END /////////////

void ff_st_end()
{
    fesetround(FE_TONEAREST);
    unsigned long long sum;
    std::cout << "----------------------------------------------------------------" << std::endl;
    std::cout << PACKAGE_STRING << " software" << std::endl;
    sum = _ff_st_div_count + _ff_st_mul_count + _ff_st_math_count + _ff_st_power_count +
          _ff_st_intrinsic_count + _ff_st_cancel_count + _ff_st_branching_count;

    if (_ff_st_div_change || _ff_st_mul_change || _ff_st_power_change)
    {
        std::cout << std::endl;
        std::cout << "BE CAREFUL: the self-validation has been deactivated." << std::endl;
        std::cout << "The results are NOT guaranteed." << std::endl;
        std::cout << std::endl;
    }

    if (!sum)
        std::cout << "No instability detected" << std::endl;
    else
    {
        if (_ff_st_div_count + _ff_st_mul_count + _ff_st_power_count)
        {
            std::cout << std::endl;
            std::cout << "CRITICAL WARNING: the self-validation detects major problem(s)." << std::endl;
            std::cout << "The results are NOT guaranteed." << std::endl;
            std::cout << std::endl;
        }

        if (sum == 1)
            std::cout << "There is  1 numerical instability" << std::endl;
        else
            std::cout << "There are " << sum << " numerical instabilities" << std::endl;

        if (_ff_st_div_tag && _ff_st_div_count)
            std::cout << _ff_st_div_count << " " << _ff_st_div_err_msg << std::endl;
        if (_ff_st_mul_tag && _ff_st_mul_count)
            std::cout << _ff_st_mul_count << " " << _ff_st_mul_err_msg << std::endl;
        if (_ff_st_power_tag && _ff_st_power_count)
            std::cout << _ff_st_power_count << " " << _ff_st_power_err_msg << std::endl;
        if (_ff_st_branching_tag && _ff_st_branching_count)
            std::cout << _ff_st_branching_count << " " << _ff_st_branching_err_msg << std::endl;
        if (_ff_st_intrinsic_tag && _ff_st_intrinsic_count)
            std::cout << _ff_st_intrinsic_count << " " << _ff_st_intrinsic_err_msg << std::endl;
        if (_ff_st_math_tag && _ff_st_math_count)
            std::cout << _ff_st_math_count << " " << _ff_st_math_err_msg << std::endl;
        if (_ff_st_cancel_tag && _ff_st_cancel_count)
            std::cout << _ff_st_cancel_count << " " << _ff_st_cancel_err_msg << std::endl;
    }
    #ifdef FF_ST_HALF_EMULATION
    if (_ff_st_half_overflow_tag && _ff_st_half_overflow_count) {
        std::cout << std::endl;
        std::cout << "BE CAREFUL: over range values have been detected." << std::endl;
        std::cout << "The results are NOT guaranteed." << std::endl ;
        std::cout << _ff_st_half_overflow_count << " " << _ff_st_half_overflow_err_msg << std::endl; 
        std::cout << std::endl;
    }
    
        if (_ff_st_half_underflow_tag && _ff_st_half_underflow_count) {
        std::cout << std::endl;
        std::cout << "BE CAREFUL: under range values have been detected." << std::endl;
        std::cout << "The results are NOT guaranteed." << std::endl ;
        std::cout << _ff_st_half_underflow_count << " " << _ff_st_half_underflow_err_msg << std::endl; 
        std::cout << std::endl;
    }
    #endif  
    std::cout << "----------------------------------------------------------------" << std::endl;
}
